# noinspection PyPep8Naming
class Deletable:
    """A property whose value may be deleted."""

    @property
    def delete(self):
        """
        Whether the client should delete existing samples or interval data for this property.

        Data will be deleted for the containing interval,
        or if there is no containing interval,
        then all data.
        If true,
        all other properties in this property
        will be ignored.
        """
        return self._delete


# noinspection PyPep8Naming
class Interpolatable:
    """A property whose value may be determined by interpolating.

    The interpolation happens over provided time-tagged samples.
    """

    @property
    def epoch(self):
        """The epoch to use for times specified as seconds since an epoch."""
        return self._epoch

    @property
    def interpolationAlgorithm(self):
        """The interpolation algorithm to use when interpolating."""
        return self._interpolation_algorithm

    @property
    def interpolationDegree(self):
        """The degree of interpolation to use when interpolating."""
        return self._interpolation_degree


# noinspection PyPep8Naming
class HasAlignment:
    """A property that can be horizontally or vertically aligned."""

    @property
    def horizontalOrigin(self):
        """The horizontal origin of the object.

        It controls whether the object is
        left-, center-, or right-aligned with the position.

        """
        return self._horizontal_origin

    @property
    def verticalOrigin(self):
        """The vertical origin of the object.

        Determines whether the object is
        bottom-, center-, or top-aligned with the position.

        """
        return self._vertical_origin



from .base import BaseCZMLObject
from .common import Deletable, HasAlignment, Interpolatable
from .enums import ClockRanges, ClockSteps, LabelStyles
from .types import Cartesian3Value, Uri


# noinspection PyPep8Naming
class Position(BaseCZMLObject, Interpolatable, Deletable):
    """Defines a position. The position can optionally vary over time."""

    KNOWN_PROPERTIES = [
        "delete",
        "epoch",
        "interpolationAlgorithm",
        "interpolationDegree",
        "referenceFrame",
        "cartesian",
    ]

    def __init__(
        self,
        *,
        delete=None,
        epoch=None,
        interpolationAlgorithm=None,
        interpolationDegree=None,
        referenceFrame=None,
        cartesian=None,
    ):
        if isinstance(cartesian, list):
            cartesian = Cartesian3Value(values=cartesian)

        self._delete = delete
        self._epoch = epoch
        self._interpolation_algorithm = interpolationAlgorithm
        self._interpolation_degree = interpolationDegree
        self._reference_frame = referenceFrame
        self._cartesian = cartesian

    @property
    def referenceFrame(self):
        """The reference frame in which cartesian positions are specified."""
        return self._reference_frame

    @property
    def cartesian(self):
        """The position specified as a three-dimensional Cartesian value.

        The value [X, Y, Z] is specified
        in meters relative to the ReferenceFrame.
        """
        return self._cartesian


# noinspection PyPep8Naming
class Billboard(BaseCZMLObject, HasAlignment):
    """A billboard, or viewport-aligned image.

    The billboard is positioned in the scene by the position property.
    A billboard is sometimes called a marker.
    """

    KNOWN_PROPERTIES = [
        "show",
        "image",
        "scale",
        "pixelOffset",
        "eyeOffset",
        "horizontalOrigin",
        "verticalOrigin",
        "heightReference",
        "color",
        "rotation",
        "alignedAxis",
        "sizeInMeters",
        "width",
        "height",
        "scaleByDistance",
        "translucencyByDistance",
        "pixelOffsetScaleByDistance",
        "imageSubRegion",
        "distanceDisplayCondition",
        "disableDepthTestDistance",
    ]

    def __init__(
        self,
        *,
        image,
        show=None,
        scale=None,
        horizontalOrigin=None,
        verticalOrigin=None,
    ):
        if isinstance(image, str):
            image = Uri(uri=image)
        self._image = image
        self._show = show
        self._scale = scale
        self._horizontal_origin = horizontalOrigin
        self._vertical_origin = verticalOrigin

    @property
    def show(self):
        """Whether or not the billboard is shown."""
        return self._show

    @property
    def image(self):
        """The URI of the image displayed on the billboard.

        For broadest client compatibility,
        the URI should be accessible via Cross-Origin Resource Sharing (CORS).
        The URI may also be a data URI.
        """
        return self._image

    @property
    def scale(self):
        """The scale of the billboard.

        The scale is multiplied with the pixel size of the billboard's image.
        For example, if the scale is 2.0,
        the billboard will be rendered with twice the number of pixels,
        in each direction, of the image.

        """
        return self._scale


# noinspection PyPep8Naming
class Clock(BaseCZMLObject):
    """Initial settings for a simulated clock when a document is loaded.

    The start and stop time are configured using the interval property.

    """

    KNOWN_PROPERTIES = ["currentTime", "multiplier", "range", "step"]

    def __init__(
        self,
        *,
        currentTime=None,
        multiplier=1.0,
        range=ClockRanges.LOOP_STOP,
        step=ClockSteps.SYSTEM_CLOCK_MULTIPLIER,
    ):
        self._current_time = currentTime
        self._multiplier = multiplier
        self._range = range
        self._step = step

    @property
    def currentTime(self):
        """The current time, specified in ISO8601 format."""
        return self._current_time

    @property
    def multiplier(self):
        """The multiplier.

        When step is set to TICK_DEPENDENT,
        this is the number of seconds to advance each tick.
        When step is set to SYSTEM_CLOCK_DEPENDENT,
        this is multiplied by the elapsed system time between ticks.
        This value is ignored in SYSTEM_CLOCK mode.

        """
        return self._multiplier

    @property
    def range(self):
        """The behavior when the current time reaches its start or end times."""
        return self._range

    @property
    def step(self):
        """How the current time advances each tick."""
        return self._step


# noinspection PyPep8Naming
class Path(BaseCZMLObject):
    """A path, which is a polyline defined by the motion of an object over time.

    The possible vertices of the path are specified by the position property.
    Note that because clients cannot render a truly infinite path,
    the path must be limited,
    either by defining availability for this object,
    or by using the leadTime and trailTime properties.

    """

    KNOWN_PROPERTIES = [
        "show",
        "leadTime",
        "trailTime",
        "width",
        "resolution",
        "material",
        "distanceDisplayCondition",
    ]

    def __init__(
        self, *, show=True, leadTime=None, trailTime=None, width=1.0, resolution=60.0
    ):
        self._show = show
        self._lead_time = leadTime
        self._trail_time = trailTime
        self._width = width
        self._resolution = resolution

    @property
    def show(self):
        """Whether or not the path is shown."""
        return self._show

    @property
    def leadTime(self):
        """The time ahead of the animation time, in seconds, to show the path.

        The time will be limited to not exceed the object's availability.
        By default, the value is unlimited,
        which effectively results in drawing the entire available path of the object.

        """
        return self._lead_time

    @property
    def trailTime(self):
        """The time behind the animation time, in seconds, to show the path.

        The time will be limited to not exceed the object's availability.
        By default, the value is unlimited,
        which effectively results in drawing the entire available path of the object.

        """
        return self._trail_time

    @property
    def width(self):
        """The width of the path line."""
        return self._width

    @property
    def resolution(self):
        """The maximum step-size, in seconds, used to sample the path.

        If the position property has data points
        farther apart than resolution specifies,
        additional samples will be computed,
        creating a smoother path.

        """
        return self._resolution


# noinspection PyPep8Naming
class Label(BaseCZMLObject, HasAlignment):
    """A string of text."""

    KNOWN_PROPERTIES = [
        "show",
        "text",
        "style",
        "scale",
        "showBackground",
        "backgroundColor",
        "backgroundPadding",
        "pixelOffset",
        "eyeOffset",
        "horizontalOrigin",
        "verticalOrigin",
        "heightReference",
        "fillColor",
        "outlineColor",
        "outlineWidth",
        "translucencyByDistance",
        "pixelOffsetScaleByDistance",
        "scaleByDistance",
        "distanceDisplayCondition",
        "disableDepthTestDistance",
    ]

    def __init__(
        self,
        *,
        show=True,
        text=None,
        style=LabelStyles.FILL,
        scale=None,
        showBackground=None,
        horizontalOrigin=None,
        verticalOrigin=None,
        outlineWidth=1.0,
    ):
        self._show = show
        self._text = text
        self._style = style
        self._scale = scale
        self._show_background = showBackground
        self._horizontal_origin = horizontalOrigin
        self._vertical_origin = verticalOrigin
        self._outline_width = outlineWidth

    @property
    def show(self):
        """Whether or not the label is shown."""
        return self._show

    @property
    def text(self):
        """The text displayed by the label.

        The newline character (\n) indicates line breaks.

        """
        return self._text

    @property
    def style(self):
        """The style of the label."""
        return self._style

    @property
    def scale(self):
        """The scale of the label.

        The scale is multiplied with the pixel size of the label's text.
        For example, if the scale is 2.0,
        the label will be rendered with twice the number of pixels,
        in each direction, of the text.

        """
        return self._scale

    @property
    def showBackground(self):
        """Whether or not a background behind the label is shown."""
        return self._show_background

    @property
    def outlineWidth(self):
        """The outline width of the label."""
        return self._outline_width



import datetime as dt

# TODO: Wait until https://github.com/python/typeshed/pull/2694 is released
from dateutil.parser import isoparse as parse_iso_date  # type: ignore
from w3lib.url import is_url, parse_data_uri

from .base import BaseCZMLObject
from .common import Deletable
from .constants import ISO8601_FORMAT_Z

TYPE_MAPPING = {bool: "boolean"}


def format_datetime_like(dt_object):
    if dt_object is None:
        result = dt_object

    elif isinstance(dt_object, str):
        try:
            parse_iso_date(dt_object)
        except Exception:
            raise
        else:
            result = dt_object

    elif isinstance(dt_object, dt.datetime):
        result = dt_object.astimezone(dt.timezone.utc).strftime(ISO8601_FORMAT_Z)

    else:
        result = dt_object.strftime(ISO8601_FORMAT_Z)

    return result


class Cartesian3Value(BaseCZMLObject):
    """A three-dimensional Cartesian value specified as [X, Y, Z].

    If the values has three elements, the value is constant.
    If it has four or more elements, they are time-tagged samples
    arranged as [Time, X, Y, Z, Time, X, Y, Z, ...],
    where Time is an ISO 8601 date and time string or seconds since epoch.

    """

    def __init__(self, *, values):
        if not (len(values) == 3 or len(values) % 4 == 0):
            raise ValueError(
                "Input values must have either 3 or N * 4 values, "
                "where N is the number of time-tagged samples."
            )

        self._values = values

    @property
    def values(self):
        return self._values

    def to_json(self):
        return list(self.values)


class StringValue(BaseCZMLObject, Deletable):
    """A string value.

    The string can optionally vary with time.
    """

    def __init__(self, *, delete=None, string=None):
        self._delete = delete
        self._string = string

    @property
    def string(self):
        """The string value."""
        return self._string

    def to_json(self):
        return self.string


class Uri(BaseCZMLObject, Deletable):
    """A URI value.

    The URI can optionally vary with time.
    """

    def __init__(self, *, delete=None, uri=None):
        try:
            parse_data_uri(uri)
        except ValueError as e:
            if not is_url(uri):
                raise ValueError("uri must be a URL or a data URI") from e

        self._delete = delete
        self._uri = uri

    @property
    def uri(self):
        """The URI value."""
        return self._uri

    def to_json(self):
        return self.uri


class TimeInterval(BaseCZMLObject):
    """A time interval, specified in ISO8601 interval format."""

    def __init__(self, *, start=None, end=None):
        self._start = format_datetime_like(start)
        self._end = format_datetime_like(end)

    def to_json(self):
        if self._start is None:
            start = "0000-00-00T00:00:00Z"
        else:
            start = self._start

        if self._end is None:
            end = "9999-12-31T24:00:00Z"
        else:
            end = self._end

        return "{start}/{end}".format(start=start, end=end)


class IntervalValue(BaseCZMLObject):
    """Value over some interval."""

    def __init__(self, *, start, end, value):
        self._interval = TimeInterval(start=start, end=end)
        self._value = value

    def to_json(self):
        obj_dict = {"interval": self._interval}

        try:
            obj_dict.update(**self._value.to_json())
        except AttributeError:
            key = TYPE_MAPPING[type(self._value)]
            obj_dict[key] = self._value

        return obj_dict


class Sequence(BaseCZMLObject):
    """Sequence, list, array of objects."""

    def __init__(self, values):
        self._values = values

    def to_json(self):
        return list(self._values)



from uuid import uuid4

from .base import BaseCZMLObject
from .types import Sequence

CZML_VERSION = "1.0"


class Preamble(BaseCZMLObject):
    """The preamble packet."""

    KNOWN_PROPERTIES = ["id", "version", "name", "clock"]

    def __init__(self, *, version=CZML_VERSION, name=None, clock=None):
        self._id = "document"
        self._version = version
        self._name = name
        self._clock = clock

    @property
    def id(self):
        return self._id

    @property
    def version(self):
        """The CZML version being written."""
        return self._version

    @property
    def name(self):
        return self._name

    @property
    def clock(self):
        """The clock settings for the entire data set."""
        return self._clock


class Packet(BaseCZMLObject):
    """A CZML Packet.

    See https://github.com/AnalyticalGraphicsInc/czml-writer/wiki/Packet
    for further information.
    """

    # https://github.com/AnalyticalGraphicsInc/czml-writer/wiki/Packet
    KNOWN_PROPERTIES = [
        "id",
        "delete",
        "name",
        "parent",
        "description",
        "availability",
        "properties",
        "position",
        "orientation",
        "viewFrom",
        "billboard",
        "box",
        "corridor",
        "cylinder",
        "ellipse",
        "ellipsoid",
        "label",
        "model",
        "path",
        "polygon",
        "polyline",
        "rectangle",
        "wall",
    ]

    def __init__(
        self,
        *,
        id=None,
        delete=None,
        name=None,
        parent=None,
        description=None,
        availability=None,
        properties=None,
        position=None,
        billboard=None,
        label=None,
        path=None,
    ):
        if id is None:
            id = str(uuid4())

        self._id = id
        self._delete = delete
        self._name = name
        self._parent = parent
        self._description = description
        self._availability = availability
        self._properties = properties
        self._position = position
        self._billboard = billboard
        self._label = label
        self._path = path

    @property
    def id(self):
        """The ID of the object described by this packet.

        IDs do not need to be GUIDs,
        but they do need to uniquely identify a single object within a CZML source
        and any other CZML sources loaded into the same scope.
        If this property is not specified,
        the client will automatically generate a unique one.
        However,
        this prevents later packets from referring to this object
        in order to add more data to it.
        """
        return self._id

    @property
    def delete(self):
        """Whether the client should delete all existing data for this object.

        If true, all other properties in this packet will be ignored.
        """
        return self._delete

    @property
    def name(self):
        """The name of the object.

        It does not have to be unique and is intended for user consumption.
        """
        return self._name

    @property
    def parent(self):
        """An HTML description of the object."""
        return self._parent

    @property
    def description(self):
        """An HTML description of the object."""
        return self._description

    @property
    def availability(self):
        """The set of time intervals over which data for an object is available.

        The property can be a single string specifying a single interval,
        or an array of strings representing intervals.
        A later CZML packet can update this availability
        if it changes or is found to be incorrect.
        For example, an SGP4 propagator may initially report availability for all time,
        but then later the propagator throws an exception
        and the availability can be adjusted to end at that time.
        If this optional property is not present,
        the object is assumed to be available for all time.
        Availability is scoped to a particular CZML stream,
        so two different streams can list different availability for a single object.
        Within a single stream,
        the last availability stated for an object is the one in effect
        and any availabilities in previous packets are ignored.
        If an object is not available at a time,
        the client will not draw that object.
        """
        return self._availability

    @property
    def properties(self):
        """A set of custom properties for this object."""
        return self._properties

    @property
    def position(self):
        """The position of the object in the world.

        The position has no direct visual representation,
        but it is used to locate billboards, labels,
        and other graphical items attached to the object.
        """
        return self._position

    @property
    def billboard(self):
        """A billboard, or viewport-aligned image, sometimes called a marker.

        The billboard is positioned in the scene by the position property.

        """
        return self._billboard

    @property
    def label(self):
        """A string of text.

        The label is positioned in the scene by the position property.

        """
        return self._label

    @property
    def path(self):
        """A path, which is a polyline defined by the motion of an object over time.

        The possible vertices of the path are specified by the position property.

        """
        return self._path


class Document(Sequence):
    """A CZML document, consisting on a list of packets."""

    @property
    def packets(self):
        return self._values
